/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Nagappan A <nagappan@gmail.com>
 *    S Thanikachalam <sthanikachalam@novell.com>
 *    Premkumar J <jpremkumar@novell.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "ldtp.h"
#include "ldtp-gui.h"
#include "ldtp-utils.h"
#include "ldtp-error.h"
#include "ldtp-logger.h"
#include "ldtp-appmap.h"
#include "ldtp-command.h"
#include "ldtp-gui-comp.h"

static char*
read_delimiter (int fd, char delimit)
{
	int len = 128;
	int size = 0;
	char ch;
	char *data = NULL;

	while (read (fd, &ch, 1) > 0) {
		if (!data)
			data = (char *) malloc (sizeof (char)*len);
		if (size%len == 0)
			data = (char *) realloc (data, sizeof (char)*len+size+1);
		data[size++] = ch;
		if (data[0] == '\n' || ch == delimit || (size && data[size - 1] == ']')) {
			if (data[0] != '\n' && read (fd, &ch, 1) > 0) {
				/*
				 * Don't do any operation
				 */
			}
			if (size%len == 0)
				data = (char *) realloc (data, sizeof (char)+size+1);
			data[size] = '\0';
			return data;
		}
	}
	return NULL;
}

static void
add_child_attributes (char *cur_entry, GHashTable *cur_context_table)
{
	char *token = NULL;
	int offset;
	char *component = NULL;
	    
	token = strtok (strdup (cur_entry), "=");
	if (token)
		component = strdup (token);
	    
	offset = strlen (component) + 1;
	if (cur_entry[offset] == '{') {
		char *end = strstr (cur_entry, "}");
		if (end) {
			char *key, *value;
			char *all_attributes;
			GHashTable *hash_attributes;
			int len = (end - cur_entry) - (offset+1);
	 
			all_attributes = (char *) malloc (sizeof (char) * len + 1);
			strncpy (all_attributes, (cur_entry + offset + 1), len);
			all_attributes[len] = '\0';

			/*
			  Create new attribute hash table
			*/
			hash_attributes = g_hash_table_new (&g_str_hash, &g_str_equal);
			key = strtok (strdup (all_attributes), "=");
			while (key) {
				char *hash_key;
				char *hash_value;
				hash_key = g_strdup (key);
				value = strtok (NULL, ",");
				hash_value = g_strdup (value);
				if (hash_key && hash_value)
					g_print ("%s: %s\n", hash_key, hash_value);
				if (hash_attributes && hash_key && hash_value)
					g_hash_table_insert (hash_attributes, hash_key, hash_value);
				key = strtok (NULL, "=");
			}
			if (component && cur_context_table && hash_attributes)
				g_hash_table_insert (cur_context_table, component, hash_attributes);
		}
	}
}

void
print_attributes (char *key, char *value, char *userdata)
{
	if (key && value)
		g_print ("\t|-------%s = %s\n", key, value);
}

void
print_component (char *key, GHashTable *component, char *userdata)
{
	if (key)
		g_print ("|-------%s\n", key);
	g_hash_table_foreach (component, (GHFunc)&print_attributes, userdata);
}

void
print_context (char *key, GHashTable *context, char *userdata)
{
	if (key)
		g_print (":%s:\n", key);
	g_hash_table_foreach (context, (GHFunc)&print_component, userdata);
}

/*
  Initialize application map
*/
GHashTable*
appmap_init (char *gui_map_filename, FILE *log_fp)
{
	int fd;
	//int quit_flag = 0;
	FILE *fp;
	char *cur_entry = NULL;
	GHashTable *appmap;
	GHashTable *cur_context_table = NULL;

	appmap = g_hash_table_new (&g_str_hash, &g_str_equal);
    
	fp = fopen (gui_map_filename, "r");
	if (fp == NULL) {
		g_print ("Unable to open appmap %s file\n", gui_map_filename);
		log_msg (LDTP_LOG_CAUSE, "Unable to open appmap file", log_fp);
		return NULL;
	}
	fd = fileno (fp);

	cur_entry = read_delimiter (fd, '}');
	while (1) {
		int cur_entry_len;
		char *context = NULL;
		if (!cur_entry)
			break;
		cur_entry_len = strlen (cur_entry);
		if (cur_entry [0] == '[' && cur_entry [cur_entry_len - 1] == ']') {
			cur_context_table = g_hash_table_new (&g_str_hash, &g_str_equal);
			context = (char *) g_new0 (char, sizeof (char) * (cur_entry_len - 1));

			/* Copy everything except the leading '[' and trailing ']' */
			memcpy (context, cur_entry+1, cur_entry_len-2);
			if (appmap && context && cur_context_table) 
				g_hash_table_insert (appmap, context, cur_context_table);
		}
		else if (cur_entry[0] == '\n') {
			cur_entry = read_delimiter (fd, '}');
			continue;
		}
		else
			add_child_attributes (cur_entry, cur_context_table);
		cur_entry = read_delimiter (fd, '}');
	}
	//g_hash_table_foreach (appmap, (GHFunc)&print_context, NULL);
	return appmap;
}

static gboolean
remove_component_entries (gpointer key, gpointer value, gpointer user_data)
{
	if (key)
		g_free (key);
	key = NULL;
	if (value)
		g_free (value);
	value = NULL;
	return TRUE;
}

static gboolean
remove_context_entries (gpointer key, gpointer value, gpointer context)
{
	GHashTable *component = NULL;

	component = g_hash_table_lookup (context, (char *)key);
	if (component) {
		g_hash_table_foreach_remove (component, (GHRFunc)&remove_component_entries, NULL);
		if (key)
			g_free (key);
		g_hash_table_destroy (component);
		key = NULL;
		component = NULL;
	}
	return TRUE;
}

static gboolean
remove_appmap_entries (gpointer key, gpointer value, gpointer user_data)
{
	GHashTable *context = NULL;
	GHashTable *appmap = user_data;

	context = g_hash_table_lookup (appmap, (char *)key);
	if (context) {
		g_hash_table_foreach_remove (context, (GHRFunc)&remove_context_entries, context);
		if (key)
			g_free (key);
		g_hash_table_destroy (context);
		key = NULL;
		context = NULL;
	}
	return TRUE;
}

/*
  Free application map resource
*/
void
ldtp_appmap_free (GHashTable* appmap)
{
	if (appmap) {
		g_hash_table_foreach_remove (appmap, (GHRFunc)&remove_appmap_entries, appmap);
		g_hash_table_destroy (appmap);
	}
}

static gboolean
search_key_regexp_based (gpointer key, gpointer value, gpointer user_data)
{
	if (!key && !user_data)
		return FALSE;
	/*
	  FIXME: Use get_object_info function to get the the type of object
	  if type and key matches then return true
	*/
	return g_pattern_match_string (user_data, key);
}

static gboolean
search_label_based (gpointer key, gpointer value, gpointer user_data)
{
	char *tmp = NULL;
	gchar *label = NULL;
	gchar *label_by = NULL;
	gchar *tmp_data = (gchar *)user_data;

	if (!value && !tmp_data) {
		g_print ("value or tmp_data NULL\n");
		return FALSE;
	}
	label = g_hash_table_lookup (value, "label");
	/*
	  FIXME: Use get_object_info function to get the the type of object
	  if type and key matches then return true
	*/
	if (label && tmp_data) {
		if ((gchar *)key && label && tmp_data)
			g_print ("Label: %s - %s - %s\n", (gchar *)key, (char *)tmp_data, label);
		if (g_utf8_collate (label, tmp_data) == 0)
			return TRUE;
		if (g_utf8_strchr (label, -1, '_') == NULL)
			return FALSE;
		tmp = escape_character (label, '_');
		if (tmp && tmp_data)
			if (g_utf8_collate (tmp, tmp_data) == 0)
				return TRUE;
		return FALSE;
	}
	label_by = g_hash_table_lookup (value, "label_by");
	if (label_by && tmp_data) {
		if ((gchar *)key && label_by && tmp_data)
			g_print ("LabelBy: %s - %s - %s\n", (char *)key, (char *)user_data, label_by);
		if (g_utf8_collate (label_by, tmp_data) == 0)
			return TRUE;
		if (g_utf8_strchr (label_by, -1, '_') == NULL)
			return FALSE;
		tmp = escape_character (label, '_');
		if (tmp && tmp_data)
			if (g_utf8_collate (tmp, tmp_data) == 0)
				return TRUE;
		return FALSE;
	}
	return FALSE;
}

static gboolean
search_label_regexp_based (gpointer key, gpointer value, gpointer user_data)
{
	char *tmp_label = NULL;
	char *tmp_label_by = NULL;

	if (!value && !user_data)
		return FALSE;
	/*
	  FIXME: Remove under score from label or label_by and do a pattern match too
	*/

	tmp_label = g_hash_table_lookup (value, "label");
	if (tmp_label) {
		if (key && value)
			g_print ("Label regexp: %s - %s\n", (char *)key, (char *)value);
		return g_pattern_match_string (user_data, tmp_label);
	}
	tmp_label_by = g_hash_table_lookup (value, "label_by");
	if (tmp_label_by) {
		g_print ("Label_By regexp: %s\n", tmp_label_by);
		return g_pattern_match_string (user_data, tmp_label_by);
	}
	return FALSE;
}

static gboolean 
search_label_name (gpointer key, gpointer value, gpointer user_data)
{
	
	char *child_index = NULL;
	UnknLabelProperty *tmp_data = (UnknLabelProperty *) user_data;
	if (!value && !tmp_data && !key) {
		return FALSE;
	}
	child_index = g_hash_table_lookup (value, "child_index");
	g_print("==DEBUG== search_label_name: %s - %s\n", (char *)key, child_index);
	if (child_index) {
		if ((char *)key && tmp_data->str_child_index && tmp_data->parent_name)
			g_print ("Key - Child index - Parent: %s - %s - %s\n",
				 (char *)key, tmp_data->str_child_index,
				 tmp_data->parent_name);
		if (child_index && tmp_data->str_child_index &&
		    g_utf8_collate (child_index, tmp_data->str_child_index) == 0) {
			char *parent_name = NULL;
			parent_name = g_hash_table_lookup (value, "parent");
			g_print ("Key - Parent: %s - %s\n",
				 (char *)key,
				 parent_name);
			if (parent_name && tmp_data->parent_name &&
			    g_utf8_collate (parent_name, tmp_data->parent_name) == 0) {
				tmp_data->obj_name = (char *) key;
				return TRUE;
			}
		}
 	}
	return FALSE;
}

static gboolean
search_obj_after_stripping_space (gpointer key, gpointer value, gpointer user_data)
{
	char *tmp;
	if (!key && !user_data)
		return FALSE;
	if (g_utf8_strchr (user_data, -1, ' ') == NULL)
		return FALSE;
	tmp = escape_character (user_data, ' ');
	if (!tmp)
		return FALSE;
	g_print ("DEBUG: %s - %s - %s\n", tmp, (char *)user_data, (char *)key);
	if (g_utf8_collate (key, tmp) == 0 || g_utf8_collate (key, user_data) == 0) {
		g_free (tmp);
		return TRUE;
	}
	else {
		char *regexp_str = NULL;
		/*
		  FIXME: Use get_object_info function to get the the type of object
		  and based on that use the convention, instead of directly using *
		*/
		regexp_str = g_strdup_printf ("*%s", tmp);
		if (regexp_str) {
			g_print ("Regexp str: %s\n", regexp_str);
			if (g_pattern_match_simple (regexp_str, key) ||
			    g_pattern_match_simple (key, tmp)) {
				g_free (tmp);
				g_free (regexp_str);
				return TRUE;
			}
		}
		else {
			gboolean flag = g_pattern_match_simple (key, tmp);
			g_free (tmp);
			return flag;
		}
		g_free (tmp);
	}
	return FALSE;
}

/*
  Get window definition
*/
GHashTable*
get_object_def (GHashTable *ht, char *context, FILE *log_fp)
{
	GHashTable *ht_context;
	LDTPError *err = NULL;
	char msg[256];

	if (ht == NULL) {
		err = ldtp_error_get_error (LDTP_ERROR_APPMAP_NOT_INITIALIZED);
		g_print ("%s - %d - %s\n", __FILE__, __LINE__, err->message);
		log_msg (LDTP_LOG_CAUSE, err->message, log_fp);
		return NULL;
	}
	if (context == NULL) {
		err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		g_print ("%s\n", err->message);
		log_msg (LDTP_LOG_CAUSE, err->message, log_fp);
		return NULL;
	}
	/*
	  Search given context in hash table
	*/
	ht_context = g_hash_table_lookup (ht, (gconstpointer) context);

	if (!ht_context) {
		GPatternSpec *pattern = NULL;
		/*
		  Search key based on regular expression
		*/
		g_print ("Search key based on regular expression: %s\n", context);
		pattern = g_pattern_spec_new ((const gchar *)context);

		ht_context = g_hash_table_find (ht, search_key_regexp_based, pattern);
		g_pattern_spec_free (pattern);
	}
	if (!ht_context) {
		/*
		  Search label based
		*/
		g_print ("Search label based: %s\n", context);
		ht_context = g_hash_table_find (ht, search_label_based, context);
	}
	if (!ht_context) {
		GPatternSpec *pattern = NULL;
		/*
		  Search label based on regular expression
		*/
		g_print ("Search label based on regular expression: %s\n", context);
		pattern = g_pattern_spec_new (context);

		ht_context = g_hash_table_find (ht, search_label_regexp_based, pattern);
		g_pattern_spec_free (pattern);
	}
	if (!ht_context) {
		/*
		  Search key after stripping space
		*/
		g_print ("Search key after stripping space: %s\n", context);
		ht_context = g_hash_table_find (ht, search_obj_after_stripping_space, context);
	}
	if (!ht_context) {
		g_sprintf (msg, "Object definition %s not found in appmap", context);
		g_print ("%s\n", msg);
		log_msg (LDTP_LOG_DEBUG, msg, log_fp);
	}
	return ht_context;
}

char*
get_property (GHashTable *ht, char *property, FILE *log_fp)
{
	char *value = NULL;
	char msg[256];
	LDTPError *err = NULL;

	memset (msg, 0x00, 256);
	if (ht == NULL) {
		err = ldtp_error_get_error (LDTP_ERROR_APPMAP_NOT_INITIALIZED);
		g_print ("%s - %d - %s\n", __FILE__, __LINE__, err->message);
		log_msg (LDTP_LOG_CAUSE, err->message, log_fp);
		return NULL;
	}
	if (property == NULL) {
		err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		g_print ("%s\n", err->message);
		log_msg (LDTP_LOG_CAUSE, err->message, log_fp);
		return NULL;
	}
	value = g_hash_table_lookup (ht, property);

	if (!value) {
		g_sprintf (msg, "Property %s not found in appmap", property);
		g_print ("%s\n", msg);
		log_msg (LDTP_LOG_DEBUG, msg, log_fp);
	}
	if (value)
		g_print ("Property: %s - Value: %s\n", property, value);
	return value;
}

char*
get_unknown_obj_label (GHashTable *ht, char *parent_name, int *child_index, int child_index_count, FILE *log_fp)
{
	int i;
	gboolean flag = FALSE;
	char *obj_name = NULL;
	//char *tmp_child_index = NULL;
	//char *tmp_parent_name = NULL;
	char str_child_index[256];
	GHashTable *ht_context = NULL;

	if (!ht) {
		LDTPError *err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		g_print ("%s\n", err->message);
		log_msg (LDTP_LOG_CAUSE, err->message, log_fp);
		return NULL;
	}

	UnknLabelProperty *unkn_label_prop = (UnknLabelProperty*) calloc (sizeof (UnknLabelProperty), 1);

	for (i = child_index_count - 1; i > 0; i--) {
		if (i == child_index_count - 1)
			unkn_label_prop->parent_name = parent_name;
		else
			unkn_label_prop->parent_name = unkn_label_prop->obj_name;
		g_print ("**DEBUG** Inside Loop for %d %s\n", i, unkn_label_prop->parent_name);
		sprintf (str_child_index, "%d", child_index[i-1]);
		unkn_label_prop->str_child_index = str_child_index;
		g_print ("--DEBUG-- Current child index and parent name are %s %s\n",
			 unkn_label_prop->parent_name, 
			 unkn_label_prop->str_child_index);
		ht_context = g_hash_table_find (ht, search_label_name, unkn_label_prop);
		if (!ht_context) {
			g_print ("HashTable Context NULL\n");
			flag = FALSE;
			//return;
			break; //Changing the break to return
		}
		g_print ("=========HashTable Context NOT NULL=========\n");
		flag = TRUE;
 	}
	if (flag)
		obj_name = unkn_label_prop->obj_name;
	g_free (unkn_label_prop);
	return obj_name;
}
